# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

from .abstract_filter import AbstractFilter

class FilterSaturation(AbstractFilter):
	__gtype_name__ = 'FilterSaturation'

	def __init__(self, filter_id, filters_tool, *args):
		super().__init__(filter_id, filters_tool)
		self._label, self._spinbtn = self._tool.bar.add_spinbtn( \
		                    _("Saturation"), [100, 0, 10000, 10, 30, 0], 3, '%')
		# it's [value, lower, upper, step_increment, page_increment, page_size]

	def get_preferred_minimum_width(self):
		return self._label.get_preferred_width()[0] + \
		     self._spinbtn.get_preferred_width()[0]

	def set_filter_compact(self, is_active, is_compact):
		self._label.set_visible(is_active and not is_compact)
		self._spinbtn.set_visible(is_active)

	def build_filter_op(self):
		options = {
			'percent': self._spinbtn.get_value() / 100,
		}
		return options

	def do_filter_operation(self, source_pixbuf, operation):
		self._tool.get_image().set_temp_pixbuf(source_pixbuf.copy())
		temp = self._tool.get_image().temp_pixbuf
		source_pixbuf.saturate_and_pixelate(temp, operation['percent'], False)

	############################################################################
################################################################################




# abstract_classic_tool.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo
from .abstract_tool import AbstractAbstractTool
from .optionsbar_classic import OptionsBarClassic
from .utilities_colors import utilities_gdk_rgba_to_normalized_array

class AbstractClassicTool(AbstractAbstractTool):
	__gtype_name__ = 'AbstractClassicTool'

	def __init__(self, tool_id, label, icon_name, window, **kwargs):
		super().__init__(tool_id, label, icon_name, window)
		self.menu_id = 0
		self.use_color = True
		self.use_size = True
		self.accept_selection = False

		self.tool_width = 10
		self.main_color = None
		self.secondary_color = None
		self._fallback_operator = 'over'
		self._operator = cairo.Operator.OVER
		self.x_press = 0
		self.y_press = 0
		self._use_antialias = self.load_tool_action_boolean('antialias', \
		                                                     'use-antialiasing')
		# XXX honteusement sous-performant ^

	############################################################################
	# UI implementations #######################################################

	def try_build_pane(self):
		self.pane_id = 'classic'
		self.window.options_manager.try_add_bottom_pane(self.pane_id, self)

	def build_bottom_pane(self):
		return OptionsBarClassic(self.window)

	############################################################################
	# Options ##################################################################

	def set_common_values(self, event_btn, event_x, event_y):
		self._use_antialias = self.get_option_value('antialias')
		self.tool_width = self.window.options_manager.get_tool_width()
		left_c = self.window.options_manager.get_left_color()
		right_c = self.window.options_manager.get_right_color()
		if event_btn == 1:
			self.main_color = utilities_gdk_rgba_to_normalized_array(left_c)
			self.secondary_color = utilities_gdk_rgba_to_normalized_array(right_c)
		if event_btn == 3:
			self.main_color = utilities_gdk_rgba_to_normalized_array(right_c)
			self.secondary_color = utilities_gdk_rgba_to_normalized_array(left_c)
		self.x_press = event_x
		self.y_press = event_y
		self._operator = self.window.options_manager.get_operator()[0]

	############################################################################
	# Operations common methods ################################################

	def start_tool_operation(self, operation):
		super().start_tool_operation(operation)
		context = self.get_context()
		if 'antialias' not in operation:
			antialias = cairo.Antialias.DEFAULT
		elif operation['antialias']:
			antialias = cairo.Antialias.DEFAULT
		else:
			antialias = cairo.Antialias.NONE
		context.set_antialias(antialias)
		return context

	def set_dashes_and_cap(self, cairo_context, lw, dashes_type, line_cap):
		"""Set a cairo dash pattern to the given context. The pattern we want
		is described by its name as a string (`dashes_type`), but the actual
		lengths that cairo wants depend also on the line width (`lw`) and the
		`line_cap`. The line cap is also set here."""
		cairo_context.set_line_cap(line_cap)
		dashes_descriptor = []
		if dashes_type == 'regular':
			dashes_descriptor = [2, 2]
		elif dashes_type == 'long':
			dashes_descriptor = [3, 1]
		elif dashes_type == 'dots':
			dashes_descriptor = [1, 1]
		elif dashes_type == 'alt':
			dashes_descriptor = [3, 1, 1, 1]

		i = 0
		for length in dashes_descriptor:
			if line_cap != cairo.LineCap.BUTT:
				if i % 2 == 0:
					length = length - 1
				else:
					length = length + 1
			actual_length = length * lw
			if length == 0:
				actual_length = actual_length + 1
			dashes_descriptor[i] = actual_length
			i = i + 1
		cairo_context.set_dash(dashes_descriptor)

	############################################################################
################################################################################



# tool_highlight.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo
from .tool_pencil import ToolPencil
from .abstract_classic_tool import AbstractClassicTool

class ToolHighlighter(ToolPencil):
	__gtype_name__ = 'ToolHighlighter'

	def __init__(self, window, **kwargs):
		# Context: this is the name of a tool, a thick pencil dedicated to
		# highlight text, for example in screenshots
		AbstractClassicTool.__init__(self, 'highlight', _("Highlighter"), \
		                                      'tool-highlight-symbolic', window)
		self.use_operator = False
		self._path = None
		self.add_tool_action_boolean('highlight-alpha', True)
		self.add_tool_action_boolean('highlight-rigid', True)
		self.add_tool_action_enum('highlight-bg', 'light')

	def get_editing_tips(self):
		self._bg_type = self.get_option_value('highlight-bg')
		self._force_alpha = self.get_option_value('highlight-alpha')
		self._is_rigid = self.get_option_value('highlight-rigid')

		label_options = self.label + " - "
		if self._bg_type == 'light':
			label_options += _("Dark text on light background")
			label_modifier_shift = _("Press <Shift> to temporarily highlight" + \
			                                      " on dark background instead")
		else:
			label_options += _("Light text on dark background")
			label_modifier_shift = _("Press <Shift> to temporarily highlight" + \
			                                     " on light background instead")
		if self.get_image().get_mouse_is_pressed():
			label_modifier_shift = None

		full_list = [label_options, label_modifier_shift]
		return list(filter(None, full_list))

	def get_options_label(self):
		return _("Highlighter options")

	############################################################################

	def on_press_on_area(self, event, surface, event_x, event_y):
		self.set_common_values(event.button, event_x, event_y)
		self._path = None

		self.update_modifier_state(event.state)
		if 'SHIFT' in self._modifier_keys:
			if self._bg_type == 'light':
				self._bg_type = 'dark'
			else:
				self._bg_type = 'light'

	def _add_point(self, event_x, event_y):
		cairo_context = self.get_context()
		if self._path is None:
			cairo_context.move_to(self.x_press, self.y_press)
		elif self._didnt_really_move(cairo_context, event_x, event_y):
			length = -1
			for pts in self._path:
				if pts[1] == ():
					continue
				length += 1
				# a better technique to find the length probably exists
			for index, pts in enumerate(self._path):
				if pts[1] == ():
					continue
				if pts[0] == cairo.PathDataType.MOVE_TO:
					cairo_context.move_to(pts[1][0], pts[1][1])
				elif index == length:
					event_x = (pts[1][0] + event_x) / 2
					event_y = (pts[1][1] + event_y) / 2
					break
				else: # if pts[0] == cairo.PathDataType.LINE_TO:
					cairo_context.line_to(pts[1][0], pts[1][1])
		cairo_context.line_to(event_x, event_y)
		self._path = cairo_context.copy_path()

	def _didnt_really_move(self, cairo_context, event_x, event_y):
		"""Tells if the pointer has moved enough to add a new point, otherwise
		the last point will be changed.
		It's an option that can be disabled.
		The context of an highlighter tool means the direction is biased: i will
		assume the underlying text is written horizontally, and in straight
		lines; so the highlighting will also be straight, but the chosen line
		may change during the stroke."""
		cairo_context.append_path(self._path)
		if not self._is_rigid:
			return False

		rigidity = min(self.tool_width, 10.0)
		if abs(cairo_context.get_current_point()[0] - event_x) > rigidity:
			return False
		if abs(cairo_context.get_current_point()[1] - event_y) > rigidity / 5:
			return False

		cairo_context.new_path()
		return True

	def on_motion_on_area(self, event, surface, event_x, event_y, render=True):
		self._add_point(event_x, event_y)
		if not render:
			return
		operation = self.build_operation()
		self.do_tool_operation(operation)

	def on_release_on_area(self, event, surface, event_x, event_y):
		self._add_point(event_x, event_y)
		operation = self.build_operation()
		self.apply_operation(operation)

	############################################################################

	def build_operation(self):
		operation = {
			'tool_id': self.id,
			'rgba': self.main_color,
			'width': self.tool_width,
			'path': self._path,
			'bg-type': self._bg_type,
			'halpha': self._force_alpha
		}
		return operation

	def do_tool_operation(self, operation):
		self.start_tool_operation(operation)
		if operation['path'] is None:
			return
		ccontext = self.get_context()
		ccontext.set_line_cap(cairo.LineCap.SQUARE)
		ccontext.set_line_join(cairo.LineJoin.ROUND)
		ccontext.set_line_width(operation['width'])

		if operation['bg-type'] == 'light':
			operator = cairo.Operator.MULTIPLY
		else:
			operator = cairo.Operator.SCREEN
		ccontext.set_operator(operator)

		main_color = operation['rgba']
		if operation['halpha']:
			main_color[3] = 0.5
		ccontext.set_source_rgba(*main_color)

		ccontext.append_path(operation['path'])
		ccontext.stroke()

	############################################################################
################################################################################




# tool_filters.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo
from gi.repository import Gdk, GdkPixbuf, Gio, GLib
from .abstract_transform_tool import AbstractCanvasTool
from .filter_blur import FilterBlur
from .filter_colors import FilterColors
from .filter_contrast import FilterContrast
from .filter_emboss import FilterEmboss
from .filter_saturation import FilterSaturation
from .filter_transparency import FilterTransparency
from .filter_veil import FilterVeil
from .optionsbar_filters import OptionsBarFilters
from .utilities_blur import utilities_blur_surface, BlurType, BlurDirection

class ToolFilters(AbstractCanvasTool):
	__gtype_name__ = 'ToolFilters'

	def __init__(self, window):
		super().__init__('filters', _("Filters"), 'tool-filters-symbolic', window)
		self.cursor_name = 'pointer'

		self.add_tool_action_enum('filters_type', 'saturation')
		self.type_label = _("Change saturation")
		self._active_filter = 'saturation'

		# Options specific to filters, but which are here for no good reason
		self.add_tool_action_enum('filters_blur_dir', 'none')
		self.blur_algo = BlurType.INVALID

		# Initialisation of the filters
		self._all_filters = {
			'blur': FilterBlur('blur', self),
			'colors': FilterColors('colors', self),
			'contrast': FilterContrast('contrast', self),
			'emboss': FilterEmboss('emboss', self),
			'saturation': FilterSaturation('saturation', self),
			'transparency': FilterTransparency('transparency', self),
			'veil': FilterVeil('veil', self),
		}

	def try_build_pane(self):
		self.pane_id = 'filters'
		self.window.options_manager.try_add_bottom_pane(self.pane_id, self)

	def build_bottom_pane(self):
		self.bar = OptionsBarFilters(self.window, self)
		self.bar.menu_btn.connect('notify::active', self._set_active_type)
		self.bar.menu_btn.connect('notify::active', self._set_blur_direction)
		return self.bar

	def get_max_filter_width(self):
		width = 0
		for f in self._all_filters.values():
			width = max(f.get_preferred_minimum_width(), width)
		return width

	def set_filters_compact(self, is_compact):
		for f_id, f in self._all_filters.items():
			f.set_filter_compact(f_id == self._active_filter, is_compact)

	def get_options_label(self):
		return _("Active filter")

	def get_editing_tips(self):
		tip_label = _("Click on the image to preview the selected filter")
		return [self.type_label, tip_label]

	############################################################################

	def _set_blur_direction(self, *args):
		self._all_filters['blur'].set_attributes_values()

	def _set_active_type(self, *args):
		state_as_string = self.get_option_value('filters_type')

		self.blur_algo = BlurType.INVALID
		if state_as_string == 'blur_fast':
			self.blur_algo = BlurType.CAIRO_REPAINTS
			self.type_label =  _("Fast blur")
			self._active_filter = 'blur'
		elif state_as_string == 'blur_slow':
			self.blur_algo = BlurType.PX_BOX
			self.type_label = _("Slow blur")
			self._active_filter = 'blur'
		elif state_as_string == 'tiles':
			self.blur_algo = BlurType.TILES
			self.type_label = _("Mosaic")
			self._active_filter = 'blur'

		elif state_as_string == 'saturation':
			self.type_label = _("Change saturation")
			self._active_filter = 'saturation'
		elif state_as_string == 'veil':
			self.type_label = _("Veil")
			self._active_filter = 'veil'

		elif state_as_string == 'contrast':
			self.type_label = _("Increase contrast")
			self._active_filter = 'contrast'
		# TODO changer la luminosity tant qu'à faire
		elif state_as_string == 'emboss':
			# Context: a filter. See "image embossing" on wikipedia
			self.type_label = _("Emboss")
			self._active_filter = 'emboss'

		elif state_as_string == 'invert':
			self.type_label = _("Invert colors")
			self._active_filter = 'colors'

		elif state_as_string == 'transparency':
			self.type_label = _("Add transparency")
			self._active_filter = 'transparency'
		else:
			self.type_label = _("Select a filter…")
		self.bar.on_filter_changed()

	############################################################################

	def on_tool_selected(self, *args):
		super().on_tool_selected()
		self._set_active_type()
		self._set_blur_direction()
		GLib.timeout_add(100, self._async_open_menu, {})
		if self.blur_algo == BlurType.INVALID:
			self._preview_filter()
		else:
			operation = self.build_operation()
			operation['radius'] = 0
			# when the active filter is blurring the image, we cheat a little to
			# improve performance, but the filtered pixbuf isn't preview until
			# the user explicitly clicks on the canvas to preview
			self.do_tool_operation(operation)

	def _async_open_menu(self, *args):
		"""This is used as a GSourceFunc so it should return False."""
		self.bar.menu_btn.set_active(True)
		return False

	def on_press_on_area(self, event, surface, event_x, event_y):
		self._preview_filter()

	def _preview_filter(self, *args):
		self._set_active_type()
		self._set_blur_direction()
		self.build_and_do_op()

	############################################################################

	def build_operation(self):
		operation = {
			'tool_id': self.id,
			'is_selection': self.apply_to_selection,
			'is_preview': True,
			'local_dx': 0,
			'local_dy': 0,
			'filter_id': self._active_filter
		}
		options = self._all_filters[self._active_filter].build_filter_op()
		return {**operation, **options}

	def do_tool_operation(self, operation):
		self.start_tool_operation(operation)
		if operation['is_selection']:
			source_pixbuf = self.get_selection_pixbuf()
		else:
			source_pixbuf = self.get_main_pixbuf()

		active_filter = self._all_filters[operation['filter_id']]
		active_filter.do_filter_operation(source_pixbuf, operation)

		self.common_end_operation(operation)

	############################################################################
################################################################################




# options_manager.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gio, Gdk

class DrOptionsManager():
	__gtype_name__ = 'DrOptionsManager'
	# XXX this class should raise/catch more exceptions

	_tools_gsettings = Gio.Settings.new('com.github.maoschanz.drawing.tools-options')

	def __init__(self, window):
		self.window = window
		self._bottom_panes_dict = {}
		self._active_pane_id = None
		self._boolean_actions_from_gsetting = {}
		self._string_actions_from_gsetting = {}
		self._mirror_lock = False

	def _action_exists(self, name):
		return self.window.lookup_action(name) is not None

	############################################################################
	# Gio.Action for tool options ##############################################

	def add_option_boolean(self, name, default):
		if self._action_exists(name):
			return
		self.window.add_action_boolean(name, default, self._boolean_callback)

	def add_option_enum(self, name, default):
		if self._action_exists(name):
			return
		self.window.add_action_enum(name, default, self._enum_callback)

	def add_option_from_bool_key(self, action_name, key_name):
		default_value = self._tools_gsettings.get_boolean(key_name)
		self.add_option_boolean(action_name, default_value)
		self._boolean_actions_from_gsetting[action_name] = key_name
		return default_value

	def add_option_from_enum_key(self, action_name, key_name):
		default_value = self._tools_gsettings.get_string(key_name)
		self.add_option_enum(action_name, default_value)
		self._string_actions_from_gsetting[action_name] = key_name
		return default_value

	def get_value(self, name):
		if not self._action_exists(name):
			return
		action = self.window.lookup_action(name)
		if action.get_state_type().dup_string() == 's':
			return action.get_state().get_string()
		else:
			return action.get_state()

	def _boolean_callback(self, *args):
		"""This callback is simple but can't handle both menu-items and check/
		toggle-buttons. It is only good for menu-items and model-buttons."""
		new_value = args[1].get_boolean()

		args[0].set_state(GLib.Variant.new_boolean(new_value))
		self.window.set_window_subtitles()
		self.get_active_pane().hide_options_menu()

	def _enum_callback(self, *args):
		"""This callback is simple but can't handle both menu-items and
		radio-buttons. It is only good for menu-items and model-buttons."""
		new_value = args[1].get_string()

		# No need to change the state if it's already as the user want.
		# Cases "m1 m1" or "b1 b1" or "b1 m1" or "m1 b1"
		if new_value == args[0].get_state().get_string():
			return

		# Actually change the state to the new value.
		args[0].set_state(GLib.Variant.new_string(new_value))
		self.window.set_window_subtitles()
		self.get_active_pane().hide_options_menu()

	############################################################################
	# Special case: mirrorable actions are required when using radio-buttons ###

	def mirrored_action_callback(self, mirror_action_name, *args):
		"""Actions whose callbacks use this method can be used in menus. It sets
		the lock required to avoid infinite recursion caused by the
		synchronisation with a mirror action used on GtkRadioButtons."""
		self._enum_callback(*args)
		mirroring_action = self.window.lookup_action(mirror_action_name)
		if args[1].get_string() == mirroring_action.get_state().get_string():
			return False
		self._mirror_lock = True
		self._enum_callback(mirroring_action, args[1])
		self._mirror_lock = False
		return True

	def mirroring_action_callback(self, mirrored_action_name, *args):
		"""Actions whose callbacks use this method should NEVER be added to any
		menu. Such an action should mirror the action provided as an argument
		(which can be added to menus).
		Those actions are intended to be used by GtkRadioButtons, whose weird
		behaviors include sending a `change-state` signal when being unchecked,
		thus triggering this callback twice. So a synchronisation mechanism is
		needed, with a lock & an other "classic" action duplicating the data."""
		if self._mirror_lock:
			return False
		if args[1].get_string() == args[0].get_state().get_string():
			return False
		mirrored_action = self.window.lookup_action(mirrored_action_name)
		if args[1].get_string() == mirrored_action.get_state().get_string():
			return False
		self._enum_callback(*args)
		self._mirror_lock = True
		self._enum_callback(mirrored_action, args[1])
		self._mirror_lock = False
		return True

	############################################################################

	def persist_tools_options(self, *args):
		"""Called before closing to persist the current values of a few options
		into gsettings (dconf or key-value file)."""

		# Panel-wide classic tools options (they are not Gio actions!)
		self._tools_gsettings.set_int('last-size', self.get_tool_width())
		self._persist_color(self.get_left_color(), 'last-left-rgba')
		self._persist_color(self.get_right_color(), 'last-right-rgba')

		# Tool-wide boolean actions
		for action_name in self._boolean_actions_from_gsetting:
			key_name = self._boolean_actions_from_gsetting[action_name]
			self._persist_boolean(action_name, key_name)

		# Tool-wide "enum" actions
		for action_name in self._string_actions_from_gsetting:
			key_name = self._string_actions_from_gsetting[action_name]
			self._persist_string(action_name, key_name)

	def _persist_string(self, action_name, key_name):
		action_value = self.get_value(action_name)
		self._tools_gsettings.set_string(key_name, action_value)

	def _persist_boolean(self, action_name, key_name):
		action_value = self.get_value(action_name)
		self._tools_gsettings.set_boolean(key_name, action_value)

	def _persist_color(self, rgba, key_name):
		rgba = [str(rgba.red), str(rgba.green), str(rgba.blue), str(rgba.alpha)]
		self._tools_gsettings.set_strv(key_name, rgba)

	############################################################################
	# Bottom panes management ##################################################

	def try_add_bottom_pane(self, pane_id, calling_tool):
		if pane_id not in self._bottom_panes_dict:
			new_pane = calling_tool.build_bottom_pane()
			if new_pane is None:
				return
			self._bottom_panes_dict[pane_id] = new_pane
			self.window.bottom_panes_box.add(new_pane.action_bar)

	def try_enable_pane(self, pane_id):
		if pane_id == self._active_pane_id:
			return
		elif pane_id not in self._bottom_panes_dict:
			# Shouldn't happen anyway
			return
		else:
			self._active_pane_id = pane_id
			self._show_active_pane()
			self.update_minimap_position()

	def _show_active_pane(self):
		for each_id in self._bottom_panes_dict:
			is_active = each_id == self._active_pane_id
			self._bottom_panes_dict[each_id].action_bar.set_visible(is_active)

	def get_active_pane(self):
		return self._bottom_panes_dict[self._active_pane_id]

	def update_pane(self, tool):
		self._bottom_panes_dict[tool.pane_id].update_for_new_tool(tool)

	############################################################################

	def init_adaptability(self):
		for pane_id in self._bottom_panes_dict:
			self._bottom_panes_dict[pane_id].init_adaptability()
		self._show_active_pane()

	def adapt_to_window_size(self, available_width):
		self.get_active_pane().adapt_to_window_size(available_width)
		# Because the animation to show a bottom pane takes some time, it's
		# possible to change tool "too fast" (ctrl+b really quickly) and end up
		# in a situation where no pane is visible.
		# There is no easy fix for this as far as i know.
		# The following instruction "fixes" the problem just enough so users,
		# if they encounter it, will believe they dreamt the issue.
		self.get_active_pane().action_bar.set_visible(True)

	############################################################################

	def toggle_menu(self):
		self.get_active_pane().toggle_options_menu()

	def set_minimap_label(self, label):
		for pane_id in self._bottom_panes_dict:
			self._bottom_panes_dict[pane_id].set_minimap_label(label)

	def update_minimap_position(self):
		"""Move the minimap popover to the currently active optionsbar."""
		btn = self.get_active_pane().get_minimap_btn()
		if btn is not None:
			self.window.minimap.set_relative_to(btn)
		else:
			self.window.minimap.set_relative_to(self.window.bottom_panes_box)

	############################################################################
	# Methods specific to the optionsbar for classic tools #####################

	def get_classic_tools_pane(self):
		return self._bottom_panes_dict['classic']

	def left_color_btn(self):
		return self.get_classic_tools_pane()._color_l

	def right_color_btn(self):
		return self.get_classic_tools_pane()._color_r

	def set_palette_setting(self, show_editor):
		self.get_classic_tools_pane().set_palette_setting(show_editor)

	def get_tool_width(self):
		return int(self.get_classic_tools_pane().thickness_spinbtn.get_value())

	def update_tool_width(self, delta):
		width = self.get_tool_width() + delta
		self.get_classic_tools_pane().thickness_spinbtn.set_value(width)

	def set_right_color(self, color):
		return self.right_color_btn().color_widget.set_rgba(color)

	def set_left_color(self, color):
		return self.left_color_btn().color_widget.set_rgba(color)

	def get_right_color(self):
		return self.right_color_btn().color_widget.get_rgba()

	def get_left_color(self):
		return self.left_color_btn().color_widget.get_rgba()

	def get_persisted_color(self, is_left_color):
		if is_left_color:
			color_array = self._tools_gsettings.get_strv('last-left-rgba')
		else:
			color_array = self._tools_gsettings.get_strv('last-right-rgba')
		r = float(color_array[0])
		g = float(color_array[1])
		b = float(color_array[2])
		a = float(color_array[3])
		return Gdk.RGBA(red=r, green=g, blue=b, alpha=a)

	def get_operator(self):
		# XXX répugnant, on duplique la donnée dans les 2 popovers, puis on
		# hardcode qu'on prendra la donnée uniquement dans celui de droite, et
		# tout ça sans respecter l'encapsulation
		enum = self.get_classic_tools_pane()._color_r._operator_enum
		label = self.get_classic_tools_pane()._color_r._operator_label
		return enum, label

	############################################################################
################################################################################




# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

import cairo, random
from .abstract_brush import AbstractBrush

class BrushAirbrush(AbstractBrush):
	__gtype_name__ = 'BrushAirbrush'

	def _get_tips(self, use_pressure, brush_direction):
		label = _("Airbrush") + " - "
		if use_pressure:
			label += _("Density depends on the stylus pressure")
		else:
			label += _("Density depends on the mouse speed")
		return [label]

	def do_brush_operation(self, cairo_context, operation):
		"""Airbrush whose radius is the line width. The density of droplets is
		related to the stylus pressure."""

		cairo_context.set_operator(operation['operator'])
		cairo_context.set_source_rgba(*operation['rgba'])

		# if operation['is_preview']: # Previewing helps performance & debug
		# 	operation['line_width'] = int(operation['line_width'] / 2)
		# 	return self.draw_preview(operation, cairo_context)

		cairo_context.set_line_width(1)
		random.seed(1) # this hardcoded seed avoids the droplets changing their
		# positions when the user undoes an following operation
		half_width = int(operation['line_width'] / 2)
		droplets = 20 # could be like 15 + log(width) maybe?

		for pt in operation['path']:
			if pt['p'] is not None:
				droplets = int(40 * pt['p'])
			for i in range(droplets):
				cairo_context.new_path()
				x = pt['x'] + random.randint(-1 * half_width, half_width)
				y = pt['y'] + random.randint(-1 * half_width, half_width)
				cairo_context.move_to(x, y)
				cairo_context.rel_line_to(1, 1)
				cairo_context.stroke()
		# XXX the pattern is square, not round

	############################################################################
################################################################################




# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

import cairo
from gi.repository import Gtk, Gdk, GdkPixbuf

class DrMinimap(Gtk.Popover):
	__gtype_name__ = 'DrMinimap'

	def __init__(self, window, minimap_btn, **kwargs):
		super().__init__(**kwargs)
		self._window = window
		self._preview_size = self._window.gsettings.get_int('preview-size')
		self.mini_pixbuf = GdkPixbuf.Pixbuf.new(GdkPixbuf.Colorspace.RGB, True, 8, 300, 300)
		self._mini_surface = cairo.ImageSurface(cairo.Format.ARGB32, 5, 5)

		builder = Gtk.Builder.new_from_resource( \
		                          '/com/github/maoschanz/drawing/ui/minimap.ui')
		box = builder.get_object('minimap_box')

		self._zoom_scale = builder.get_object('_zoom_scale')
		self._zoom_scale.add_mark(100.0, Gtk.PositionType.TOP, None)
		self._zoom_scale.connect('value-changed', self._update_zoom_level)

		self._minimap_area = builder.get_object('_minimap_area')
		self._minimap_area.set_size(200, 200)
		self._minimap_area.add_events(Gdk.EventMask.BUTTON_PRESS_MASK | \
		                              Gdk.EventMask.BUTTON_MOTION_MASK | \
		                              Gdk.EventMask.BUTTON_RELEASE_MASK)
		self._minimap_area.connect('draw', self._on_mm_draw)
		self._minimap_area.connect('button-press-event', self._on_mm_press)
		self._minimap_area.connect('motion-notify-event', self._on_mm_motion)
		self._minimap_area.connect('button-release-event', self._on_mm_motion)

		self.add(box)
		self.connect('closed', self._on_popover_dismissed)
		self.set_relative_to(minimap_btn)

		# TODO custom "moving hand" cursor?

	############################################################################

	def update_zoom_scale(self, value):
		self._zoom_scale.set_value(value * 100)

	def set_zoom_label(self, int_value):
		# This string displays the zoom level: %s will be replaced with a
		# number, while %% will be rendered as the symbol '%'
		zoom_label = _("%s%%") % str(int(int_value))
		self._window.options_manager.set_minimap_label(zoom_label)
		self.update_overlay()

	def update_content(self):
		"""Update the minimap's content."""
		image = self._window.get_active_image()
		self.mini_pixbuf = image.generate_mini_pixbuf(self._preview_size)
		pix_width = self.mini_pixbuf.get_width()
		pix_height = self.mini_pixbuf.get_height()
		self._minimap_area.set_size_request(pix_width, pix_height)

		self.update_overlay(True)

	def update_overlay(self, force_update=False):
		"""Update the overlay on the minimap, based on the zoom level and the
		scroll coordinates."""
		if not self.get_visible() and not force_update:
			return
		image = self._window.get_active_image()
		self._mini_surface = Gdk.cairo_surface_create_from_pixbuf( \
		                                              self.mini_pixbuf, 0, None)
		if image.get_minimap_need_overlay():
			size_ratio = image.get_minimap_ratio(self.mini_pixbuf.get_width())
			mini_x = int(image.scroll_x * size_ratio)
			mini_y = int(image.scroll_y * size_ratio)
			visible_width, visible_height = image.get_visible_size()
			# We add pixels because those "int" truncate a pixel on each side
			mini_width = int(visible_width * size_ratio) + 2
			mini_height = int(visible_height * size_ratio) + 2

			# Set up a cairo context
			mini_context = cairo.Context(self._mini_surface)
			mini_context.new_path()
			mini_context.set_line_width(1)
			mini_context.set_antialias(cairo.Antialias.NONE)
			mini_context.set_line_cap(cairo.LineCap.SQUARE)

			# Path around the visible area
			mini_context.move_to(mini_x, mini_y)
			mini_context.line_to(mini_x, mini_height + mini_y)
			mini_context.line_to(mini_width + mini_x, mini_height + mini_y)
			mini_context.line_to(mini_width + mini_x, mini_y)
			mini_context.line_to(mini_x, mini_y)

			# Path around the entire mini-surface
			pix_width = self.mini_pixbuf.get_width()
			pix_height = self.mini_pixbuf.get_height()
			mini_context.move_to(0, 0)
			mini_context.line_to(pix_width, 0)
			# We add pixels because those "int" truncate a pixel on each side
			mini_context.line_to(pix_width + 1, pix_height + 1)
			mini_context.line_to(0, pix_height)
			mini_context.close_path()

			# Fill between these 2 paths with half-transparent grey
			mini_context.set_source_rgba(0.3, 0.3, 0.3, 0.2)
			mini_context.fill_preserve()

			# Draw the paths with grey
			mini_context.set_source_rgba(0.5, 0.5, 0.5, 1.0)
			mini_context.stroke()
		self._minimap_area.queue_draw()

	############################################################################

	def _update_zoom_level(self, *args):
		zoom_value = self._zoom_scale.get_value()
		image = self._window.get_active_image()
		if image.zoom_level != zoom_value / 100:
			image.zoom_level = zoom_value / 100
			image.update()
		self.set_zoom_label(image.zoom_level * 100)

	def _on_popover_dismissed(self, *args):
		"""Callback of the 'closed' signal, updating the state of the action."""
		try:
			self.get_relative_to().set_active(False)
		except:
			pass

	def _on_mm_draw(self, area, cairo_context):
		"""Callback of the 'draw' signal, painting the area with the surface."""
		cairo_context.set_source_surface(self._mini_surface, 0, 0)
		cairo_context.paint()

	def _on_mm_press(self, area, event):
		"""Callback of the 'button-press-event' signal."""
		self._press_x = event.x
		self._press_y = event.y
		image = self._window.get_active_image()
		self._old_scroll_x = image.scroll_x
		self._old_scroll_y = image.scroll_y

	def _on_mm_motion(self, area, event):
		"""Callback of the 'button-release-event' signal (and also the
		'motion-notify-event' signal)."""
		image = self._window.get_active_image()
		image.scroll_x = self._old_scroll_x
		image.scroll_y = self._old_scroll_y
		size_ratio = image.get_minimap_ratio(self.mini_pixbuf.get_width())
		delta_x = int((event.x - self._press_x) / size_ratio)
		delta_y = int((event.y - self._press_y) / size_ratio)
		image.add_deltas(delta_x, delta_y, 1)

	############################################################################
################################################################################




# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

import cairo
from .abstract_optionsbar import AbstractOptionsBar
from .optionsbar_color_popover import OptionsBarClassicColorPopover
from .utilities_units import utilities_add_unit_to_spinbtn

################################################################################

class OptionsBarClassic(AbstractOptionsBar):
	__gtype_name__ = 'OptionsBarClassic'

	def __init__(self, window):
		super().__init__()
		self.window = window
		builder = self._build_ui('classic/optionsbar-classic.ui')

		self.color_box = builder.get_object('color_box')
		self.color_menu_btn_r = builder.get_object('color_menu_btn_r')
		self.color_menu_btn_l = builder.get_object('color_menu_btn_l')
		self._build_color_buttons(builder)

		window.add_action_enum('cairo_operator', 'over', self._cairo_op_changed)
		window.add_action_enum('cairo_op_mirror', 'over', self._cairop_mirror)
		self._cairo_operator_lock = False

		self.options_label = builder.get_object('options_label')
		self.options_long_box = builder.get_object('options_long_box')
		self.options_short_box = builder.get_object('options_short_box')

		self.thickness_scalebtn = builder.get_object('thickness_scalebtn')
		self.thickness_spinbtn = builder.get_object('thickness_spinbtn')
		last_size = self._get_tool_options().get_int('last-size')
		self.thickness_spinbtn.set_value(last_size)
		utilities_add_unit_to_spinbtn(self.thickness_spinbtn, 3, 'px')

		self.minimap_btn = builder.get_object('minimap_btn')
		self.minimap_label = builder.get_object('minimap_label')
		self.minimap_arrow = builder.get_object('minimap_arrow')

	def _get_tool_options(self):
		return self.window.options_manager._tools_gsettings

	def update_for_new_tool(self, tool):
		self.color_box.set_sensitive(tool.use_color)
		if tool.use_operator:
			operator = self.window.options_manager.get_value('cairo_operator')
		else:
			operator = tool._fallback_operator
		self._color_r.set_operators_available(tool.use_operator, operator)
		self._color_l.set_operators_available(tool.use_operator, operator)
		self.thickness_scalebtn.set_sensitive(tool.use_size)
		self.thickness_spinbtn.set_sensitive(tool.use_size)

	def get_minimap_btn(self):
		return self.minimap_btn

	def set_minimap_label(self, label):
		self.minimap_label.set_label(label)

	def build_options_menu(self, widget, model, label):
		if widget is not None:
			self.options_btn.set_popover(widget)
		elif model is not None:
			self.options_btn.set_menu_model(model)
		else:
			self.options_btn.set_popover(None)
		self.options_label.set_label(label)

	def init_adaptability(self):
		super().init_adaptability()
		temp_limit_size = self.color_box.get_preferred_width()[0] + \
		          self.thickness_spinbtn.get_preferred_width()[0] + \
		           self.options_long_box.get_preferred_width()[0] + \
		                self.minimap_btn.get_preferred_width()[0] + 50
		# assuming 50px is enough to compensate the length of the label
		self._set_limit_size(temp_limit_size)

	def set_compact(self, state):
		super().set_compact(state)
		self.options_long_box.set_visible(not state)
		self.options_short_box.set_visible(state)
		self.thickness_scalebtn.set_visible(state)
		self.thickness_spinbtn.set_visible(not state)
		self.minimap_arrow.set_visible(not state)

	############################################################################
	# Colors ###################################################################

	def middle_click_action(self):
		left_color = self._color_l.color_widget.get_rgba()
		self._color_l.color_widget.set_rgba(self._color_r.color_widget.get_rgba())
		self._color_r.color_widget.set_rgba(left_color)

	def _build_color_buttons(self, builder):
		"""Initialize the 2 color-buttons and popovers with the 2 previously
		memorized RGBA values."""
		options_manager = self.window.options_manager

		thumbnail_r = builder.get_object('r_btn_image')
		self._color_r = OptionsBarClassicColorPopover(self.color_menu_btn_r, \
		                                    thumbnail_r, False, options_manager)

		thumbnail_l = builder.get_object('l_btn_image')
		self._color_l = OptionsBarClassicColorPopover(self.color_menu_btn_l, \
		                                     thumbnail_l, True, options_manager)

	def set_palette_setting(self, show_editor):
		self._color_r.editor_setting_changed(show_editor)
		self._color_l.editor_setting_changed(show_editor)

	############################################################################
	# Cairo operators ("color application modes") ##############################

	def _cairo_op_changed(self, *args):
		"""This action can be used in menus. Is sync'd with `cairo_op_mirror`."""
		if self.window.options_manager.mirrored_action_callback('cairo_op_mirror', *args):
			self._update_popovers(args[1].get_string())

	def _cairop_mirror(self, *args):
		"""This action should NEVER be added to any menu (GtkRadioButtons only).
		Is sync'd with `cairo_operator`."""
		if self.window.options_manager.mirroring_action_callback('cairo_operator', *args):
			self._update_popovers(args[1].get_string())

	def _update_popovers(self, op_as_string):
		self._color_r.adapt_to_operator(op_as_string)
		self._color_l.adapt_to_operator(op_as_string)

	############################################################################
################################################################################




# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

import cairo, math
from gi.repository import Gdk, GdkPixbuf
from .selection_manager import NoSelectionPathException

################################################################################
# Selection overlay ############################################################

def utilities_show_overlay_on_context(ccontext, cpath, thickness=1):
	"""Draw a blueish area on `ccontext`, for the selection."""
	if cpath is None:
		raise NoSelectionPathException()
	ccontext.new_path()
	ccontext.set_antialias(cairo.Antialias.NONE)
	ccontext.set_line_width(thickness)
	ccontext.set_dash([thickness * 3, thickness * 3])
	ccontext.append_path(cpath)
	ccontext.set_source_rgba(0.1, 0.1, 0.3, 0.2)
	ccontext.fill_preserve()
	ccontext.set_source_rgba(0.5, 0.5, 0.5, 0.5)
	ccontext.stroke()

################################################################################
# Transform tools overlay ######################################################

def utilities_show_handles_on_context(cairo_context, x1, x2, y1, y2, thickness=1):
	"""Request the drawing of handles for a rectangle pixbuf having the provided
	coords. Handles are only decorative objects drawn on the surface to help the
	user understand the rationale of tools without relying on the mouse cursor."""
	radius = _get_radius((x2 - x1), (y2 - y1), thickness)
	lateral_handles = True # may become a parameter later
	if radius < 4 * thickness:
		cairo_context.set_line_width(thickness)
	elif radius < 8 * thickness:
		cairo_context.set_line_width(2 * thickness)
	else:
		cairo_context.set_line_width(3 * thickness)

	_draw_arc_handle(cairo_context, x1, y1, radius, 'nw')
	_draw_arc_handle(cairo_context, x2, y1, radius, 'ne')
	_draw_arc_handle(cairo_context, x2, y2, radius, 'se')
	_draw_arc_handle(cairo_context, x1, y2, radius, 'sw')
	if lateral_handles:
		_draw_arc_handle(cairo_context, (x1+x2)/2, y1, radius, 'n')
		_draw_arc_handle(cairo_context, x2, (y1+y2)/2, radius, 'e')
		_draw_arc_handle(cairo_context, (x1+x2)/2, y2, radius, 's')
		_draw_arc_handle(cairo_context, x1, (y1+y2)/2, radius, 'w')

	cairo_context.move_to(x1, y1)
	cairo_context.line_to(x1, y2)
	cairo_context.line_to(x2, y2)
	cairo_context.line_to(x2, y1)
	cairo_context.close_path()

	cairo_context.set_line_width(thickness)
	cairo_context.set_dash([2 * thickness, 2 * thickness])
	cairo_context.set_source_rgba(0.5, 0.5, 0.5, 0.5)
	cairo_context.stroke()

def _draw_arc_handle(cairo_context, x, y, radius, orientation):
	"""Draw a moon-like shape with the given orientation and radius. The coords
	are the center of the shape. The orientation is an enumeration provided as a
	string with the stupid n/nw/w/sw/s/se/e/ne format."""
	if orientation == 'nw':
		angle_1 = 0.5 * math.pi
		angle_2 = 2.0 * math.pi
	elif orientation == 'n':
		angle_1 = math.pi
		angle_2 = 0.0
	elif orientation == 'ne':
		angle_1 = math.pi
		angle_2 = 0.5 * math.pi
	elif orientation == 'e':
		angle_1 = -0.5 * math.pi
		angle_2 = 0.5 * math.pi
	elif orientation == 'se':
		angle_1 = -0.5 * math.pi
		angle_2 = math.pi
	elif orientation == 's':
		angle_1 = 0.0
		angle_2 = math.pi
	elif orientation == 'sw':
		angle_1 = 0.0
		angle_2 = -0.5 * math.pi
	elif orientation == 'w':
		angle_1 = 0.5 * math.pi
		angle_2 = -0.5 * math.pi

	cairo_context.move_to(x, y)
	cairo_context.arc(x, y, radius, angle_1, angle_2)
	cairo_context.close_path()
	cairo_context.set_source_rgba(1.0, 1.0, 1.0, 1.0)
	cairo_context.fill_preserve()
	cairo_context.set_source_rgba(0.5, 0.5, 0.5, 0.5)
	cairo_context.stroke()

def _get_radius(dx, dy, thickness):
	return min([dx/5, dy/5, 12 * thickness])

################################################################################
# Canvas generic ouline ########################################################

def utilities_generic_canvas_outline(cairo_context, zoom_level, w, h):
	cairo_context.set_source_rgba(0.0, 0.0, 0.0, 1.0)
	cairo_context.set_dash([])
	size = max(1, int(1 / zoom_level))
	cairo_context.set_line_width(size)
	cairo_context.move_to(w + size, 0)
	cairo_context.rel_line_to(0, h + size)
	cairo_context.line_to(0, h + size)
	cairo_context.stroke_preserve()

	cairo_context.set_source_rgba(1.0, 1.0, 1.0, 1.0)
	cairo_context.set_dash([2 * size, 2 * size])
	cairo_context.stroke()

################################################################################




# message_dialog.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk

class DrMessageDialog(Gtk.MessageDialog):
	__gtype_name__ = 'DrMessageDialog'

	def __init__(self, window, **kwargs):
		super().__init__(modal=True, transient_for=window, **kwargs)
		self.set_resizable(True)
		self.set_default_size(350, -1)
		self._min_action_index = 0

		# The dialog has already a default empty label, with methods around it,
		# but i don't care i'll add my widgets myself later.
		self.get_message_area().get_children()[0].destroy()

		if window.get_allocated_width() < 500:
			self.get_action_area().set_orientation(Gtk.Orientation.VERTICAL)
			self.should_wrap = True
		else:
			# Wrappable labels bother the height allocation, so they shouldn't
			# be used if it's not required
			self.should_wrap = False

	def set_action(self, label, style, is_default=False):
		"""Add a button (with the specified style) to the message dialog and
		returns the corresponding response id."""
		self._min_action_index = self._min_action_index + 1
		btn = self.add_button(label, self._min_action_index)
		if style is not None:
			btn.get_style_context().add_class(style)
		if is_default:
			btn.grab_default()
		return self._min_action_index

	def add_string(self, string):
		label = Gtk.Label(label=string, wrap=self.should_wrap)
		self.get_message_area().add(label)
		self.get_message_area().show_all()

	def add_widget(self, widget):
		self.get_content_area().add(widget)
		self.get_content_area().show_all()

	############################################################################
################################################################################



# abstract_optionsbar.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk

RSRC_PREFIX = '/com/github/maoschanz/drawing/optionsbars/'

class AbstractOptionsBar():
	__gtype_name__ = 'AbstractOptionsBar'

	def __init__(self):
		# Quite high as a precaution, will be more precise later
		self._limit_size = 700
		self._is_narrow = False

	def _build_ui(self, end_of_path):
		"""Creates a Gtk.Builder for the resource whose path is hinted in the
		method parameter, and tries to extract usual widgets ids from it. Most
		resulting widgets can be `None`, except `action_bar`. The builder is
		returned if the caller wants to extract more objects from it.
		This is valid for any instance of `AbstractOptionsBar`."""
		builder = Gtk.Builder.new_from_resource(RSRC_PREFIX + end_of_path)
		self.action_bar = builder.get_object('bottom-pane')
		self.cancel_btn = builder.get_object('cancel_btn') # may be None
		self.centered_box = builder.get_object('centered_box') # may be None
		self.apply_btn = builder.get_object('apply_btn') # may be None
		self.help_btn = builder.get_object('help_btn') # may be None
		self.options_btn = builder.get_object('options_btn') # may be None
		self._togglable_btn = self.options_btn # default value, may change later
		return builder # for implementations-specific widgets

	def _hydrate_transform_tool(self, end_of_path):
		"""Same as above, BUT this isn't valid for any instance of
		`AbstractOptionsBar`: only an optionsbar for a transform tool should
		call that method, and it should call it *after* `_build_ui`. This method
		will "hydrate" the previously built (but still incomplete) widgets."""
		builder = Gtk.Builder.new_from_resource(RSRC_PREFIX + end_of_path)
		self.centered_box = builder.get_object('centered_box')
		self.action_bar.set_center_widget(self.centered_box)
		actions_menu = builder.get_object('actions-menu')
		if actions_menu:
			self.options_btn.set_menu_model(actions_menu)
		else:
			self.options_btn.destroy()
			self.options_btn = None
		return builder

	def build_options_menu(self, widget, model, label):
		"""Set a widget (or a menu) as the popover with tools options. In
		practice this is used only with the classic tools' pane, and all classic
		tools provide a menu instead of a widget."""
		pass

	def update_for_new_tool(self, tool):
		"""Update the pane when a tool using it has just been activated. Widgets
		may be hidden/shown or enabled/disabled depending on the properties of
		the given tool."""
		pass

	def get_minimap_btn(self):
		pass

	def set_minimap_label(self, label):
		pass

	def toggle_options_menu(self):
		if self._togglable_btn and self._togglable_btn.get_visible():
			self._togglable_btn.set_active(not self._togglable_btn.get_active())

	def hide_options_menu(self):
		if self._togglable_btn:
			self._togglable_btn.set_active(False)

	def middle_click_action(self):
		pass

	def init_adaptability(self):
		"""Instructions to run during the first window size allocation (and only
		the first). Calling this is managed by the window's `DrOptionsManager`.
		It estimates the minimal amount of pixels for the mobile version, and
		gives it to the `_set_limit_size` method.
		Any implementation HAS TO end with a call to `_set_limit_size`!"""
		self.set_compact(False)
		self.action_bar.show_all()
		# + implementation-specific instructions

	def _set_limit_size(self, temp_limit_size):
		self._limit_size = int(1.25 * temp_limit_size)
		self.set_compact(True)

	def adapt_to_window_size(self, allocated_width):
		"""Check whether or not the costly `set_compact` method should be
		called, depending on a given window width, and the pane's limit
		previously set."""
		can_expand = (allocated_width > self._limit_size)
		incoherent = (can_expand == self._is_narrow)
		if incoherent:
			self.set_compact(not self._is_narrow)

	def set_compact(self, state):
		"""The parameter is a boolean telling if the bottom pane should become
		compact or not."""
		self._is_narrow = state
		if self.help_btn is not None:
			self.help_btn.set_visible(not state)
		# + implementation-specific instructions

	############################################################################
################################################################################




# tool_pencil.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo, math
from .abstract_classic_tool import AbstractClassicTool
from .utilities_paths import utilities_smooth_path

class ToolPencil(AbstractClassicTool):
	__gtype_name__ = 'ToolPencil'

	def __init__(self, window, **kwargs):
		super().__init__('pencil', _("Pencil"), 'tool-pencil-symbolic', window)
		self.use_operator = True

		self._path = None
		self._shape_label = _("Round")
		self._cap_id = cairo.LineCap.ROUND
		self._join_id = cairo.LineCap.ROUND
		self._dashes_type = 'none'
		self._use_outline = False

		self.add_tool_action_enum('line_shape', 'round')
		self.add_tool_action_enum('dashes-type', self._dashes_type)
		self.add_tool_action_boolean('pencil-outline', self._use_outline)

	def _set_active_shape(self, *args):
		state_as_string = self.get_option_value('line_shape')
		if state_as_string == 'thin':
			self._cap_id = cairo.LineCap.BUTT
			self._join_id = cairo.LineJoin.BEVEL
			self._shape_label = _("Square")
		else:
			self._cap_id = cairo.LineCap.ROUND
			self._join_id = cairo.LineJoin.ROUND
			self._shape_label = _("Round")

	def get_options_label(self):
		return _("Pencil options")

	def get_editing_tips(self):
		self._dashes_type = self.get_option_value('dashes-type')
		use_dashes = self._dashes_type != 'none'
		self._use_outline = self.get_option_value('pencil-outline')
		self._set_active_shape()

		if use_dashes:
			label_options = self.label + " - " + _("Dashed")
		else:
			label_options = None

		if self.get_image().get_mouse_is_pressed():
			label_modifier_alt = None
			if label_options is None:
				label_options = self.label
		else:
			label_modifier_alt = self.label + " - " + \
			                     _("Press <Alt> to toggle the 'outline' option")

		full_list = [label_options, label_modifier_alt]
		return list(filter(None, full_list))

	############################################################################

	def on_press_on_area(self, event, surface, event_x, event_y):
		self.set_common_values(event.button, event_x, event_y)
		self._path = None

		self.update_modifier_state(event.state)
		if 'ALT' in self._modifier_keys:
			self._use_outline = not self._use_outline

	def _add_point(self, event_x, event_y):
		cairo_context = self.get_context()
		if self._path is None:
			cairo_context.move_to(self.x_press, self.y_press)
		else:
			cairo_context.append_path(self._path)
		cairo_context.line_to(event_x, event_y)
		self._path = cairo_context.copy_path()

	def on_motion_on_area(self, event, surface, event_x, event_y, render=True):
		self._add_point(event_x, event_y)
		if not render:
			return
		operation = self.build_operation()
		self.do_tool_operation(operation)

	def on_release_on_area(self, event, surface, event_x, event_y):
		self._add_point(event_x, event_y)

		if self.x_press == event_x and self.y_press == event_y:
			# Special case when the pointer hasn't moved: in order to "force"
			# cairo to draw a pixel, a tiny segment is added artifically
			float_x, float_y = self.get_image().get_event_coords(event, False)
			decimals_x = float_x - event_x
			if decimals_x > 0:
				delta_x = 0.20
			else:
				delta_x = -0.20
			decimals_y = float_y - event_y
			if decimals_y > 0:
				delta_y = 0.20
			else:
				delta_y = -0.20
			self._add_point(event_x + delta_x, event_y + delta_y)

		operation = self.build_operation()
		self.apply_operation(operation)

	############################################################################

	def build_operation(self):
		operation = {
			'tool_id': self.id,
			'rgba': self.main_color,
			'rgba2': self.secondary_color,
			'antialias': self._use_antialias,
			'smooth': not self.get_image().is_zoomed_surface_sharp(),
			'outline': self._use_outline,
			'operator': self._operator,
			'line_width': self.tool_width,
			'line_cap': self._cap_id,
			'line_join': self._join_id,
			'dashes': self._dashes_type,
			'path': self._path
		}
		return operation

	def do_tool_operation(self, operation):
		if operation['path'] is None:
			return
		cairo_context = self.start_tool_operation(operation)

		cairo_context.set_operator(operation['operator'])
		line_width = operation['line_width']
		self.set_dashes_and_cap(cairo_context, line_width, \
		                        operation['dashes'], operation['line_cap'])
		cairo_context.set_line_join(operation['line_join']) # XXX useless?

		if operation['smooth']:
			utilities_smooth_path(cairo_context, operation['path'])
		else:
			cairo_context.append_path(operation['path'])

		if operation['outline']:
			cairo_context.set_source_rgba(*operation['rgba2'])
			cairo_context.set_line_width(line_width * 1.2 + 2)
			cairo_context.stroke_preserve()

		cairo_context.set_source_rgba(*operation['rgba'])
		cairo_context.set_line_width(line_width)
		cairo_context.stroke()

	############################################################################
################################################################################

